# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Definition of a masked array.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
"""
import warnings

import numpy as np


def _merge_masks(ma1, ma2):
    """Merge the masks of two MadArray objects and return the arguments used
    for initialisation of the resulting MadArray.

    Parameters
    ----------
    ma1 : MadArray
       First masked array to consider.
    ma2 : MadArray
       Second masked array to consider.

    Returns
    -------
    dict
        Arguments to be used for the initialisation of a MadArray.
    """
    if ma1._complex_masking or ma2._complex_masking:

        if ma1._complex_masking:
            mm1 = ma1.unknown_magnitude_mask
            mp1 = ma1.unknown_phase_mask
        else:
            mm1 = ma1.unknown_mask
            mp1 = ma1.unknown_mask

        if ma2._complex_masking:
            mm2 = ma2.unknown_magnitude_mask
            mp2 = ma2.unknown_phase_mask
        else:
            mm2 = ma2.unknown_mask
            mp2 = ma2.unknown_mask

        mask_magnitude = np.logical_or(mm1, mm2)
        mask_phase = np.logical_or(mp1, mp2)

        return {'mask_magnitude': mask_magnitude, 'mask_phase': mask_phase}
    else:
        return {'mask': np.logical_or(ma1.unknown_mask, ma2.unknown_mask)}


def _complex_masking_only(f):
    def decorated(self):
        if not self._complex_masking:
            errmsg = 'Method not defined if masking is not complex.'
            raise ValueError(errmsg)
        return f(self)
    return decorated


class MadArray(np.ndarray):
    """Subclass of numpy.ndarray to handle data with missing elements.

    .. _type_entry_madarray:

    **Type of entry**: entries of array can be *int*, *float*, or *complex*.

    .. _masking_madarray:

    **Masking**: the masking of entries has two different modes:

    * entries can be either masked or not masked, leading to a boolean mask,
      whose entries are equal to True if the corresponding data entry is
      masked, or False otherwise. This is the default mode;
    * complex entries can have only the magnitude or phase component masked, or
      both. The resulting mask has integers entries, equal to:

        * *0* if the phase and the magnitude are not masked (known magnitude
          and phase);
        * *1* if only the phase is masked (known magnitude, unknown phase);
        * *2* if only the magnitude is masked (unknown magnitude, known phase);
        * *3* if the magnitude and the phase are masked (unknown magnitude and
          phase).

      This mode is selected by setting :paramref:`complex_masking` to True.
      Entries are converted to a complex type.

    .. _indexing_madarray:

    **Indexing**: two different mode to index a MadArray are implemented:

    * a MadArray array of shape corresponding to the indices is returned, with
      both the data matrix and the mask properly indexed. This is the default
      mode;
    * a MadArray with the shape unchanged is returned, where non-indexed
      entries are set as masked. This mode is selected by setting the parameter
      :paramref:`masked_indexing` to True.

    .. _numpy_behaviour_madarray:

    **Numpy behaviour**: it is possible to use standard operations (+, -, /,
    //, \*, T) between two :class:`MadArray`, likewise operations between numpy
    arrays. The resulting object has a mask consisting of the union of the
    operands. It is also possible to use pickle operations to jointly store the
    data and the mask.

    Parameters
    ----------
    data : array_like
        One-dimensional array. See :ref:`Type of Entry<type_entry_madarray>`.
    mask : boolean array_like, optional
        Mask if type of entry is int or float. See
        :ref:`Masking<masking_madarray>`.
    mask_magnitude : boolean array_like or None, optional
        Magnitude mask if type of entry is complex. See
        :ref:`Masking<masking_madarray>`.
    mask_phase : boolean or array_like or None, optional
        Phase mask if type of entry is complex. See
        :ref:`Masking<masking_madarray>`.
    complex_masking : bool or None, optional
        Indicate how the masking is performed. If None, set to False. See
        :ref:`Masking<masking_madarray>`.
    masked_indexing : bool or None, optional
        Indicate how the indexing is performed. If None, set to False. See
        :ref:`Indexing<indexing_madarray>`.

    Warnings
    --------
    This class inherits from ndarray or subclass of ndarray. Instances can be
    then manipulated like ndarrays (e.g., indexation). While some methods have
    been implemented taking into account the mask, some may cause unexpected
    behavior (e.g., mean).

    See also
    --------
    :mod:`numpy.doc.subclassing`.

    Notes
    -----
    This class implements an alternative masked array different from
    np.ma.MadArray. The reason of this choice is that it is only used as a
    container of a ndarray and a mask. No masked operations are needed.
    """

    def __new__(cls, data, mask=None, mask_magnitude=None, mask_phase=None,
                masked_indexing=None, complex_masking=None, **kwargs):

        _data = np.array(data, **kwargs)

        if complex_masking is None:
            if isinstance(data, MadArray):
                complex_masking = data._complex_masking
            else:
                complex_masking = False

        if masked_indexing is None:
            if isinstance(data, MadArray):
                masked_indexing = data._masked_indexing
            else:
                masked_indexing = False

        if not (np.issubdtype(_data.dtype, np.floating) or
                np.issubdtype(_data.dtype, np.integer) or
                np.issubdtype(_data.dtype, np.complexfloating)):
            errmsg = 'Invalid dtype: {}'
            raise TypeError(errmsg.format(data.dtype))

        if not complex_masking:

            if mask_magnitude is not None:
                warnmsg = 'Argument `mask_magnitude` is ignored.'
                warnings.warn(warnmsg)

            if mask_phase is not None:
                warnmsg = 'Argument `mask_phase` is ignored.'
                warnings.warn(warnmsg)

            if mask is None:
                if isinstance(data, MadArray):
                    mask = data.unknown_mask
                else:
                    mask = np.zeros(_data.shape, dtype=np.bool)
            else:
                mask = np.array(mask, dtype=np.bool)
                if mask.shape != _data.shape:
                    errmsg = "Mask shape {} and data shape {} not compatible."
                    raise ValueError(errmsg.format(mask.shape, data.shape))
                mask = mask

        else:

            if not np.issubdtype(_data.dtype, np.complexfloating):
                _data = _data.astype(np.complex)

            if mask is not None:
                warnmsg = 'Argument `mask` is ignored.'
                warnings.warn(warnmsg)

            if mask_magnitude is None:

                if isinstance(data, MadArray):
                    if data._complex_masking:
                        mask_magnitude = data.unknown_magnitude_mask
                    else:
                        mask_magnitude = data.unknown_mask
                else:
                    mask_magnitude = np.zeros_like(data, dtype=np.bool)
            else:
                if mask_magnitude.shape != _data.shape:
                    errmsg = 'Magnitude mask shape {} and data shape {} not '\
                        'compatible.'
                    raise ValueError(errmsg.format(mask_magnitude.shape,
                                                   _data.shape))

            if mask_phase is None:

                if isinstance(data, MadArray):
                    if data._complex_masking:
                        mask_phase = data.unknown_phase_mask
                    else:
                        mask_phase = data.unknown_mask
                else:
                    mask_phase = np.zeros_like(data, dtype=np.bool)
            else:
                if mask_phase.shape != _data.shape:
                    errmsg = 'Phase mask shape {} and data shape {} not '\
                        'compatible.'
                    raise ValueError(errmsg.format(
                        mask_phase.shape, _data.shape))

            mask = np.zeros(_data.shape, dtype=np.uint8)
            mask[np.logical_and(mask_phase, ~mask_magnitude)] = 1
            mask[np.logical_and(~mask_phase, mask_magnitude)] = 2
            mask[np.logical_and(mask_phase, mask_magnitude)] = 3

        # create the object
        obj = np.ndarray.__new__(cls, _data.shape, dtype=_data.dtype)
        obj[:] = _data
        obj._mask = mask
        obj._masked_indexing = masked_indexing
        obj._complex_masking = complex_masking

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._mask = getattr(obj, '_mask', None)
        self._complex_masking = getattr(obj, '_complex_masking', None)
        self._masked_indexing = getattr(obj, '_masked_indexing', None)

    def __array_wrap__(self, obj, context=None):
        return obj[()] if obj.shape == () else obj

    def __getitem__(self, index):

        if (getattr(self, '_masked_indexing', None) is not None and
                self._masked_indexing):

            mask = np.zeros(self.shape, dtype=np.bool)
            mask[index] = True
            return MadArray(self, mask=np.logical_or(~mask, self._mask))
        else:
            out_arr = super().__getitem__(index)
            if getattr(out_arr, '_mask', None) is not None:
                out_arr._mask = out_arr._mask[index]
            return out_arr

    def __reduce__(self):
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._mask, self._complex_masking,
                                        self._masked_indexing)
        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._mask = state[-3]
        self._complex_masking = state[-2]
        self._masked_indexing = state[-1]
        super().__setstate__(state[0:-3])

    @property
    def known_mask(self):
        """Boolean mask for known coefficients.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if the corresponding value
            in the object is known. If complex masking, entries set to True if
            both magnitude and phase are known.
        """
        if self._complex_masking:
            return self._mask == 0
        else:
            return ~self._mask

    @property
    def unknown_mask(self):
        """Boolean mask for unknown coefficients.

        Returns
        -------
        boolean nd-array
            Boolean array with values set to True if the corresponding value in
            the object is unknown. If complex masking, entries set to True if
            magnitude and/or phase are unknown.
        """
        if self._complex_masking:
            return self.any_unknown_mask
        else:
            return np.copy(self._mask)

    @property
    @_complex_masking_only
    def any_unknown_mask(self):
        """Boolean mask for coefficients with unknown magnitude and/or phase.

        Not defined if non complex masking.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if magnitude and/or phase
            are unknown.

        Raises
        ------
        ValueError
            If `complex_masking` is False.
        """
        return ~(self._mask == 0)

    @property
    @_complex_masking_only
    def all_unknown_mask(self):
        """Boolean mask for coefficients with both unknown magnitude and phase.

        Not defined if non complex masking.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if both magnitude and phase
            are unknown.

        Raises
        ------
        ValueError
            If `complex_masking` is False.
        """
        return self._mask == 3

    @property
    @_complex_masking_only
    def known_phase_mask(self):
        """Boolean mask for coefficients with known phase.

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if phase is known.

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return ~np.logical_or(self._mask == 1, self._mask == 3)

    @property
    @_complex_masking_only
    def unknown_phase_mask(self):
        """Boolean mask for coefficients with unknown phase.

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if phase is unknown

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return np.logical_or(self._mask == 1, self._mask == 3)

    @property
    @_complex_masking_only
    def known_magnitude_mask(self):
        """Boolean mask for coefficients with known magnitude.

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if magnitude is known.

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return ~np.logical_or(self._mask == 2, self._mask == 3)

    @property
    @_complex_masking_only
    def unknown_magnitude_mask(self):
        """Boolean mask for coefficients with unknown magnitude

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if magnitude is unknown.

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return np.logical_or(self._mask == 2, self._mask == 3)

    @property
    @_complex_masking_only
    def unknown_phase_only_mask(self):
        """Boolean mask for coefficients with unknown phase and known magnitude.

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if phase is unknown and
            magnitude is known.

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return self._mask == 1

    @property
    @_complex_masking_only
    def unknown_magnitude_only_mask(self):
        """Boolean mask for coefficients with unknown magnitude and known phase.

        Not defined for int or float entries.

        Returns
        -------
        boolean nd-array
            Boolean array with entries set to True if magnitude is unknown and
            phase is known.

        Raises
        ------
        ValueError
            If masking is not complex.
        """
        return self._mask == 2

    @property
    def n_missing_data(self):
        """Number of missing data.

        Returns
        -------
        double or tuple
            Number of masked coefficients if dtype is int or float. Number of
            masked coefficients in phase and magnitude masks if dtype is
            complex.

        """
        if self._complex_masking:
            return (np.sum(self.unknown_magnitude_mask),
                    np.sum(self.unknown_phase_mask))
        else:
            return np.sum(self.unknown_mask)

    @property
    def ratio_missing_data(self):
        """Ratio of missing data.

        Returns
        -------
        double or tuple
            Ratio of masked coefficients if dtype is int or float. Ratio of
            masked coefficients in phase and magnitude masks if dtype is
            complex.

        """
        if self._complex_masking:
            return (np.average(self.unknown_magnitude_mask),
                    np.average(self.unknown_phase_mask))
        else:
            return np.average(self.unknown_mask)

    def is_masked(self):
        """Indicate if one or several elements are masked."""
        return np.any(self._mask)

    def to_np_array(self, fill_value=None):
        """Return a numpy array.

        If :paramref:`fill_value` is not None, masked elements are replaced
        according to the type of entries:

        * :paramref:`fill_value` if the type of entries is *int* or *float*;
        * If the type is *complex*, missing entries are replaced either by:
            * a complex number with the known magnitude value without the phase
              information if only the phase is masked;
            * a complex number of magnitude 1 with the known phase if only the
              magnitude is masked;
            * by :paramref:`fill_value` if both magnitude and phase are masked.

        Parameters
        ----------
        fill_value : scalar or None
            Value used to fill masked elements. If None, the initial value is
            kept.

        Returns
        -------
        nd-array
        """
        data = np.array(self)
        if fill_value is not None:
            if self._complex_masking:
                upom = self.unknown_phase_only_mask
                umom = self.unknown_magnitude_only_mask
                data[upom] = np.abs(data[upom])
                data[umom] = np.exp(1j * np.angle(data[umom]))
                data[self.all_unknown_mask] = fill_value
            else:
                data[self.unknown_mask] = fill_value
        return data

    def __add__(self, other):
        if isinstance(other, MadArray):
            if self._complex_masking or other._complex_masking:
                errmsg = 'Operation not permitted when complex masking.'
                raise ValueError(errmsg)

            return MadArray(np.add(self.to_np_array(),
                                   other.to_np_array()),
                            **_merge_masks(self, other),
                            masked_indexing=self._masked_indexing)
        else:
            return super().__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __mul__(self, other):
        if isinstance(other, MadArray):
            if self._complex_masking or other._complex_masking:
                errmsg = 'Operation not permitted when complex masking.'
                raise ValueError(errmsg)

            return MadArray(np.multiply(self.to_np_array(),
                                        other.to_np_array()),
                            **_merge_masks(self, other),
                            masked_indexing=self._masked_indexing)
        else:
            return super().__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, MadArray):
            if self._complex_masking or other._complex_masking:
                errmsg = 'Operation not permitted when complex masking.'
                raise ValueError(errmsg)

            return MadArray(np.true_divide(self.to_np_array(),
                                           other.to_np_array()),
                            **_merge_masks(self, other),
                            masked_indexing=self._masked_indexing)
        else:
            return super().__truediv__(other)

    def __floordiv__(self, other):
        if isinstance(other, MadArray):
            if self._complex_masking or other._complex_masking:
                errmsg = 'Operation not permitted when complex masking.'
                raise ValueError(errmsg)

            return MadArray(np.floor_divide(self.to_np_array(),
                                            other.to_np_array()),
                            **_merge_masks(self, other),
                            masked_indexing=self._masked_indexing)
        else:
            return super().__floordiv__(other)

    def __eq__(self, other):
        if isinstance(other, MadArray):
            return np.logical_and(self.to_np_array(0) == other.to_np_array(0),
                                  self._mask == self._mask)
        else:
            return np.array(self) == other

    def __ne__(self, other):
        return np.logical_not(self == other)

    def is_equal(self, other):
        if not isinstance(other, MadArray):
            return False

        if not np.all(self == other):
            return False

        if not (self._complex_masking == other._complex_masking and
                self._masked_indexing == other._masked_indexing):
            return False

        return True

    @property
    def T(self):
        """Transpose of the MadArray."""
        return self.transpose()

    def copy(self):
        return MadArray(self)

    def transpose(self):
        mat = super().transpose()
        mat._mask = mat._mask.transpose()
        return mat

    def __str__(self):
        arr = np.array(self)

        if np.issubdtype(self.dtype, np.integer):
            arr = arr.astype(np.float64)

        arr[self.unknown_mask] = np.nan
        arr_str = np.ndarray.__str__(arr)
        arr_str = arr_str.replace('nan', 'x')

        if np.issubdtype(self.dtype, np.integer):
            arr_str = arr_str.replace('.', '')

        string = 'MadArray, dtype={}, {} missing entries ({:.1f}%)\n{}'

        return string.format(self.dtype,
                             self.n_missing_data,
                             100 * self.ratio_missing_data, arr_str)

    def __repr__(self):
        string = '<MadArray at {}>'
        return string.format(hex(id(self)))



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Test of the module :mod:`mad_array`.

NOTE: This module uses random generation for some tests, the use of
pytest-randomly is recommanded to manage the seeds used in these tests.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
"""
import tempfile
import pytest
import pickle

import numpy as np

from madarrays.mad_array import MadArray
from madarrays.mad_array import _merge_masks

from .utils import generate_mask_50


@pytest.fixture(scope='class')
def get_data(request):
    # Dimension and shape of data
    n_dims = np.random.randint(1, 3)
    request.cls.shape = 2 * np.random.randint(20, 50, n_dims)

    # For each dtype, generate two sets of data
    request.cls.x_float = np.random.random(request.cls.shape)
    request.cls.x_int = (np.random.random(
        request.cls.shape) * 2**20).astype(np.int) + 1
    request.cls.x_complex = np.random.random(
        request.cls.shape) + 1j * np.random.random(request.cls.shape)

    request.cls.x2_float = np.random.random(request.cls.shape)
    request.cls.x2_int = (np.random.random(
        request.cls.shape) * 2**20).astype(np.int) + 1
    request.cls.x2_complex = np.random.random(
        request.cls.shape) + 1j * np.random.random(request.cls.shape)

    # Generate two sets of tree masks with 50% of missing data
    request.cls.m = generate_mask_50(request.cls.shape)
    request.cls.mp = generate_mask_50(request.cls.shape)
    request.cls.mm = generate_mask_50(request.cls.shape)

    request.cls.m2 = generate_mask_50(request.cls.shape)
    request.cls.mp2 = generate_mask_50(request.cls.shape)
    request.cls.mm2 = generate_mask_50(request.cls.shape)

    # List of indices for the indexation
    request.cls.indexes = [slice(item, np.random.randint(item + 1, dim_size))
                           for dim_size in request.cls.shape
                           for item in [np.random.randint(0, dim_size - 1)]]

    request.cls.empty_mask = np.zeros(request.cls.shape, dtype=np.bool)
    request.cls.full_mask = np.ones(request.cls.shape, dtype=np.bool)


@pytest.mark.usefixtures('get_data')
class TestMadArray:

    def test_init_unmasked(self):

        for x in [self.x_float, self.x_int, self.x_complex]:
            ma = MadArray(x)

            np.testing.assert_equal(ma.to_np_array(), x)
            assert ma.dtype == x.dtype
            assert not ma._complex_masking
            assert not ma._masked_indexing
            np.testing.assert_equal(ma.unknown_mask, self.empty_mask)
            np.testing.assert_equal(ma.known_mask, self.full_mask)
            assert ma.ratio_missing_data == 0.0
            assert ma.n_missing_data == 0
            assert not ma.is_masked()

            match = 'Method not defined if masking is not complex'
            with pytest.raises(ValueError, match=match):
                ma.any_unknown_mask
            with pytest.raises(ValueError, match=match):
                ma.all_unknown_mask
            with pytest.raises(ValueError, match=match):
                ma.known_phase_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_phase_mask
            with pytest.raises(ValueError, match=match):
                ma.known_magnitude_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_magnitude_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_phase_only_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_magnitude_only_mask

    def test_init_unmasked_complex_masking(self):

        for x in [self.x_float, self.x_int, self.x_complex]:
            ma = MadArray(x, complex_masking=True)

            np.testing.assert_equal(ma.to_np_array(), x)
            assert ma.dtype == np.complex
            assert ma._complex_masking
            assert not ma._masked_indexing
            np.testing.assert_equal(ma.unknown_mask, self.empty_mask)
            np.testing.assert_equal(ma.known_mask, self.full_mask)
            np.testing.assert_equal(ma.all_unknown_mask, self.empty_mask)
            np.testing.assert_equal(ma.known_phase_mask, self.full_mask)
            np.testing.assert_equal(ma.unknown_phase_mask, self.empty_mask)
            np.testing.assert_equal(ma.known_magnitude_mask, self.full_mask)
            np.testing.assert_equal(ma.unknown_magnitude_mask, self.empty_mask)
            np.testing.assert_equal(
                ma.unknown_phase_only_mask, self.empty_mask)
            np.testing.assert_equal(
                ma.unknown_magnitude_only_mask, self.empty_mask)
            assert ma.ratio_missing_data, (0.0, 0.0)
            assert ma.n_missing_data, (0, 0)
            assert not ma.is_masked()

    def test_init_masked(self):

        for x in [self.x_float, self.x_int, self.x_complex]:
            ma = MadArray(x, mask=self.m, complex_masking=False)

            assert ma.dtype == x.dtype
            assert not ma._complex_masking
            assert not ma._masked_indexing
            assert ma.is_masked()
            np.testing.assert_equal(ma.to_np_array(), x)
            np.testing.assert_equal(ma.unknown_mask, self.m)
            np.testing.assert_equal(ma.known_mask, ~self.m)
            assert ma.ratio_missing_data, 0.5
            assert ma.n_missing_data, x.size//2
            assert ma.is_masked()
            assert not id(ma._mask) == id(self.m)

            match = 'Method not defined if masking is not complex'
            with pytest.raises(ValueError, match=match):
                ma.any_unknown_mask
            with pytest.raises(ValueError, match=match):
                ma.all_unknown_mask
            with pytest.raises(ValueError, match=match):
                ma.known_phase_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_phase_mask
            with pytest.raises(ValueError, match=match):
                ma.known_magnitude_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_magnitude_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_phase_only_mask
            with pytest.raises(ValueError, match=match):
                ma.unknown_magnitude_only_mask

        bad_shape = [item + np.random.randint(1, 100) for item in self.shape]

        match = 'Mask shape \(\d+,( \d+)?\) and data shape \(\d+,( \d+)?\) '\
                'not compatible.'
        with pytest.raises(ValueError, match=match):
            MadArray(self.x_float, np.random.random(bad_shape) < 0.3)

        with pytest.warns(UserWarning,
                          match='Argument `mask_phase` is ignored.'):
            MadArray(x, mask=self.m, mask_phase=self.mp)

        with pytest.warns(UserWarning,
                          match='Argument `mask_magnitude` is ignored.'):
            MadArray(x, mask=self.m, mask_magnitude=self.mm)

    def test_init_masked_complex_masking(self):

        for x in [self.x_float, self.x_int, self.x_complex]:

            # With phase and magnitude masks
            ma = MadArray(x, mask_phase=self.mp, mask_magnitude=self.mm,
                          complex_masking=True)

            assert ma.dtype == np.complex
            assert ma._complex_masking
            assert not ma._masked_indexing
            assert ma.is_masked()
            np.testing.assert_equal(ma.to_np_array(), x)

            np.testing.assert_equal(ma.known_mask,
                                    np.logical_and(~self.mm, ~self.mp))
            np.testing.assert_equal(ma.unknown_mask,
                                    np.logical_or(self.mm, self.mp))
            np.testing.assert_equal(ma.any_unknown_mask,
                                    np.logical_or(self.mm, self.mp))
            np.testing.assert_equal(ma.all_unknown_mask,
                                    np.logical_and(self.mm, self.mp))
            np.testing.assert_equal(ma.known_phase_mask, ~self.mp)
            np.testing.assert_equal(ma.unknown_phase_mask, self.mp)
            np.testing.assert_equal(ma.known_magnitude_mask, ~self.mm)
            np.testing.assert_equal(ma.unknown_magnitude_mask, self.mm)
            np.testing.assert_equal(ma.unknown_phase_only_mask,
                                    np.logical_and(self.mp, ~self.mm))
            np.testing.assert_equal(ma.unknown_magnitude_only_mask,
                                    np.logical_and(~self.mp, self.mm))
            assert ma.ratio_missing_data == (0.5, 0.5)
            assert ma.n_missing_data == (x.size//2, x.size//2)

            # With magnitude mask only
            ma = MadArray(x, mask_magnitude=self.mm, complex_masking=True)

            assert ma.dtype == np.complex
            assert ma._complex_masking
            assert not ma._masked_indexing
            assert ma.is_masked()
            np.testing.assert_equal(ma.to_np_array(), x)

            np.testing.assert_equal(ma.known_mask, ~self.mm)
            np.testing.assert_equal(ma.unknown_mask, self.mm)
            np.testing.assert_equal(ma.any_unknown_mask, self.mm)
            np.testing.assert_equal(ma.all_unknown_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.known_phase_mask,
                                    np.ones(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.unknown_phase_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.known_magnitude_mask, ~self.mm)
            np.testing.assert_equal(ma.unknown_magnitude_mask, self.mm)
            np.testing.assert_equal(ma.unknown_phase_only_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.unknown_magnitude_only_mask, self.mm)
            assert ma.ratio_missing_data == (0.5, 0)
            assert ma.n_missing_data == (x.size//2, 0)

            # With mask_phase only
            ma = MadArray(x, mask_phase=self.mp, complex_masking=True)

            assert ma.dtype == np.complex
            assert ma._complex_masking
            assert not ma._masked_indexing
            assert ma.is_masked()
            np.testing.assert_equal(ma.to_np_array(), x)

            np.testing.assert_equal(ma.known_mask, ~self.mp)
            np.testing.assert_equal(ma.unknown_mask, self.mp)
            np.testing.assert_equal(ma.any_unknown_mask, self.mp)
            np.testing.assert_equal(ma.all_unknown_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.known_phase_mask, ~self.mp)
            np.testing.assert_equal(ma.unknown_phase_mask, self.mp)
            np.testing.assert_equal(ma.known_magnitude_mask,
                                    np.ones(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.unknown_magnitude_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            np.testing.assert_equal(ma.unknown_phase_only_mask, self.mp)
            np.testing.assert_equal(ma.unknown_magnitude_only_mask,
                                    np.zeros(self.shape, dtype=np.bool))
            assert ma.ratio_missing_data == (0, 0.5)
            assert ma.n_missing_data == (0, x.size//2)

        bad_shape = [item + np.random.randint(1, 100) for item in self.shape]

        match = 'Phase mask shape \(\d+,( \d+)?\) and data shape '\
                '\(\d+,( \d+)?\) not compatible.'
        with pytest.raises(ValueError, match=match):
            MadArray(self.x_complex, complex_masking=True,
                     mask_phase=np.random.random(bad_shape) < 0.8)

        match = 'Magnitude mask shape \(\d+,( \d+)?\) and data shape '\
                '\(\d+,( \d+)?\) not compatible.'
        with pytest.raises(ValueError, match=match):
            MadArray(self.x_complex, complex_masking=True,
                     mask_magnitude=np.random.random(bad_shape) < 0.8)

        with pytest.warns(UserWarning, match='Argument `mask` is ignored.'):
            MadArray(self.x_complex, mask=self.m, mask_phase=self.mp,
                     complex_masking=True)

    def test_init_maskedarray(self):

        for x in [self.x_float, self.x_int, self.x_complex]:

            old_ma = MadArray(x, self.m)
            ma = MadArray(old_ma)

            np.testing.assert_equal(ma.unknown_mask, self.m)
            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)

            old_ma = MadArray(x, self.m)
            ma = MadArray(old_ma, mask=self.mp)

            np.testing.assert_equal(ma.unknown_mask, self.mp)
            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)

    def test_init_maskedarray_complexmasking(self):

        for x in [self.x_float, self.x_int, self.x_complex]:
            old_ma = MadArray(x, mask_phase=self.mp,
                              mask_magnitude=self.mm, complex_masking=True)
            ma = MadArray(old_ma)

            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)

            old_ma = MadArray(x, mask_phase=self.mp,
                              mask_magnitude=self.mm, complex_masking=True)
            ma = MadArray(old_ma, mask_phase=self.mm,
                          mask_magnitude=self.mp)

            np.testing.assert_equal(ma.unknown_phase_mask, self.mm)
            np.testing.assert_equal(ma.unknown_magnitude_mask, self.mp)
            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)

        for x in [self.x_float, self.x_int, self.x_complex]:
            old_ma = MadArray(x, mask_phase=self.mp,
                              mask_magnitude=self.mm, complex_masking=True)

            old_ma = MadArray(x, mask=self.m, complex_masking=False)
            ma = MadArray(old_ma, complex_masking=True)

            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)
            assert ma._complex_masking

            old_ma = MadArray(x, mask_phase=self.mp,
                              mask_magnitude=self.mm, complex_masking=True)
            ma = MadArray(old_ma, mask_phase=self.mm,
                          mask_magnitude=self.mp)

            np.testing.assert_equal(ma.unknown_phase_mask, self.mm)
            np.testing.assert_equal(ma.unknown_magnitude_mask, self.mp)
            assert id(old_ma) != id(ma)
            assert id(old_ma._mask) != id(ma._mask)

    def test_init_other_dtype(self):

        # From list
        ma = MadArray(self.x_float.tolist())

        np.testing.assert_equal(ma.to_np_array(), self.x_float)
        assert np.issubdtype(ma.dtype, np.floating)

        # Raises
        x = np.array([[1, 'b'], [2, 2], [2, 4]])
        with pytest.raises(TypeError, match='Invalid dtype: <U21'):
            MadArray(x)

    def test_masked_indexing(self):

        ma = MadArray(self.x_float, self.m)
        ma_slice = ma[self.indexes]

        assert isinstance(ma_slice, MadArray)
        np.testing.assert_equal(ma_slice, self.x_float[self.indexes])
        np.testing.assert_equal(ma_slice.unknown_mask, self.m[self.indexes])
        np.testing.assert_equal(ma_slice.known_mask, ~self.m[self.indexes])
        assert ma_slice.ratio_missing_data == np.mean(self.m[self.indexes])
        assert ma_slice.n_missing_data == np.sum(self.m[self.indexes])
        assert ma.is_masked()

    def test_masked_indexing_masked(self):

        ma = MadArray(self.x_float, mask=self.m, masked_indexing=True)
        m_index = np.zeros(self.shape, dtype=np.bool)
        m_index[self.indexes] = 1

        ma_slice = ma[self.indexes]

        assert isinstance(ma_slice, MadArray)
        np.testing.assert_equal(ma_slice.shape, self.shape)
        np.testing.assert_equal(ma_slice, self.x_float)
        np.testing.assert_equal(ma_slice.unknown_mask,
                                np.logical_or(~m_index, self.m))
        np.testing.assert_equal(ma_slice.known_mask,
                                np.logical_and(m_index, ~self.m))
        assert (ma_slice.ratio_missing_data == np.mean(
            np.logical_or(~m_index, self.m)))
        assert ma_slice.n_missing_data == np.sum(
            np.logical_or(~m_index, self.m))
        assert ma.is_masked()

    def test_to_np_array(self):
        ma = MadArray(self.x_float, self.m)

        az = ma.to_np_array(fill_value=24)
        assert type(az) == np.ndarray
        np.testing.assert_equal(az[self.m], 24)
        np.testing.assert_equal(az[~self.m], self.x_float[~self.m])

        ma = MadArray(self.x_float, mask_phase=self.mp,
                      mask_magnitude=self.mm, complex_masking=True)

        xma = ma.to_np_array(0)

        np.testing.assert_almost_equal(
            xma[ma.unknown_phase_only_mask],
            np.abs(self.x_float[ma.unknown_phase_only_mask]) + 0 * 1j)
        np.testing.assert_almost_equal(
            xma[ma.unknown_magnitude_only_mask],
            np.exp(
                1j * np.angle(self.x_float[ma.unknown_magnitude_only_mask])))
        np.testing.assert_equal(xma[ma.all_unknown_mask], 0 + 0j)

        ma = MadArray(self.x_float, self.m)
        xma = ma.to_np_array()
        np.testing.assert_equal(xma, self.x_float)

    def test_copy(self):

        ma = MadArray(self.x_float, self.m)

        ma_copy = ma.copy()
        assert ma.is_equal(ma_copy)
        assert id(ma) != id(ma_copy)
        assert id(ma._mask) != id(ma_copy._mask)

    def test_operations(self):

        match = 'Operation not permitted when complex masking.'
        for operator in ['+', '-', '*', '/', '//']:
            print('Operation: {}'.format(operator))
            for x in [self.x_float, self.x_int, self.x_complex]:

                ma = MadArray(x, self.m)

                for x2 in [self.x2_float, self.x2_int, self.x2_complex]:
                    xs = eval('x {} x2'.format(operator))
                    ms = eval('ma {} x2'.format(operator))

                    assert isinstance(ms, MadArray)
                    assert not ms._complex_masking
                    assert ms.dtype == xs.dtype
                    np.testing.assert_equal(ms, xs)
                    np.testing.assert_equal(ms._mask, self.m)

                    ma2 = MadArray(x2, self.m2)
                    ms = eval('ma {} ma2'.format(operator))

                    assert isinstance(ms, MadArray)
                    assert not ms._complex_masking
                    assert ms.dtype == xs.dtype
                    np.testing.assert_equal(ms, xs)
                    np.testing.assert_equal(ms._mask,
                                            np.logical_or(self.m, self.m2))

                for x2 in [self.x2_float, self.x2_int, self.x2_complex]:

                    ma2 = MadArray(x2, mask_phase=self.mp2,
                                   mask_magnitude=self.mm2,
                                   complex_masking=True)

                    with pytest.raises(ValueError, match=match):
                        eval('ma {} ma2'.format(operator))

            for x in [self.x_float, self.x_int, self.x_complex]:
                ma = MadArray(x, mask_phase=self.mp,
                              mask_magnitude=self.mm, complex_masking=True)

                for x2 in [self.x2_float, self.x2_int, self.x2_complex]:
                    ma2 = MadArray(x2, self.m2)
                    with pytest.raises(ValueError, match=match):
                        eval('ma {} ma2'.format(operator))

                for x2 in [self.x2_float, self.x2_int, self.x2_complex]:
                    ma2 = MadArray(x2, mask_phase=self.mp2,
                                   mask_magnitude=self.mm2,
                                   complex_masking=True)

                    with pytest.raises(ValueError, match=match):
                        ma + ma2

    def test_eq_ne_numpy(self):

        ma = MadArray(self.x_float)
        eq = ma == self.x_float
        assert type(eq) == np.ndarray
        assert eq.dtype == np.bool
        assert np.all(eq)

        ma_neq = ma != self.x2_float
        assert type(ma_neq) == np.ndarray
        assert ma_neq.dtype == np.bool
        assert np.all(ma_neq)

        ma = MadArray(self.x_float, self.m)
        ma_eq = ma == self.x_float
        assert type(ma_eq) == np.ndarray
        assert ma_eq.dtype == np.bool
        assert np.all(ma_eq[ma.known_mask])

        neq = ma != self.x2_float
        assert type(neq) == np.ndarray
        assert neq.dtype == np.bool
        assert np.all(neq)

    def test_eq_ne_maskedarray(self):

        ma1 = MadArray(self.x_float, self.m)
        ma2 = MadArray(self.x_float, self.m)
        ma3 = MadArray(self.x2_float, self.m2)

        eq = ma1 == ma2
        assert type(eq) == np.ndarray
        assert eq.dtype == np.bool
        assert np.all(eq)

        neq = ma1 != ma3
        assert type(neq) == np.ndarray
        assert neq.dtype == np.bool
        assert np.any(neq)

    def test_is_equal(self):

        ma1 = MadArray(self.x_float, self.m)
        ma2 = MadArray(self.x_float, self.m)
        ma3 = MadArray(self.x_float, self.m2)
        ma4 = MadArray(self.x2_float, self.m)
        ma5 = MadArray(self.x_float, mask_phase=self.mp,
                       complex_masking=True)
        ma6 = MadArray(self.x_float, self.m, masked_indexing=True)

        assert ma1.is_equal(ma2)
        assert not ma1.is_equal(self.x_float)
        assert not ma1.is_equal(ma3)
        assert not ma1.is_equal(ma4)
        assert not ma1.is_equal(ma5)
        assert not ma1.is_equal(ma6)

    def test_transpose(self):

        ma_float = MadArray(self.x_float, self.m)

        tma = ma_float.T

        np.testing.assert_equal(tma.unknown_mask, self.m.T)
        np.testing.assert_equal(np.array(tma)[tma.known_mask],
                                self.x_float.T[tma.known_mask])

    def test_pickle(self):

        # float dtype
        ma = MadArray(self.x_float, self.m)

        with tempfile.NamedTemporaryFile() as tmp_file:
            with open(tmp_file.name, 'wb') as fout:
                pickle.dump(ma, fout)

            with open(tmp_file.name, 'rb') as fin:
                pma = pickle.load(fin)

            assert ma.is_equal(pma)

        # int  dtype
        ma = MadArray(self.x_int, self.m)

        with tempfile.NamedTemporaryFile() as tmp_file:
            with open(tmp_file.name, 'wb') as fout:
                pickle.dump(ma, fout)

            with open(tmp_file.name, 'rb') as fin:
                pma = pickle.load(fin)

            assert ma.is_equal(pma)

        # complex dtype
        ma = MadArray(self.x_complex, mask_phase=self.mp,
                      mask_magnitude=self.mm, complex_masking=True)

        with tempfile.NamedTemporaryFile() as tmp_file:
            with open(tmp_file.name, 'wb') as fout:
                pickle.dump(ma, fout)

            with open(tmp_file.name, 'rb') as fin:
                pma = pickle.load(fin)

            assert ma.is_equal(pma)

    def test_str_repr(self):

        x = np.copy(self.x_float)
        arr_str = np.ndarray.__str__(x)

        ma = MadArray(self.x_float)
        string = 'MadArray, dtype=float64, 0 missing entries (0.0%)\n{}'
        assert str(ma) == string.format(arr_str)

        x[self.m] = np.nan
        arr_str = np.ndarray.__str__(x)
        arr_str = arr_str.replace('nan', 'x')

        ma = MadArray(self.x_float, self.m)
        n_miss = np.count_nonzero(self.m)
        string = 'MadArray, dtype=float64, {} missing entries (50.0%)\n{}'
        assert str(ma) == string.format(n_miss, arr_str)

        string = '<MadArray at {}>'
        assert repr(ma) == string.format(hex(id(ma)))

        x = np.copy(self.x_int).astype(np.float64)
        x[self.m] = np.nan
        arr_str = np.ndarray.__str__(x)
        arr_str = arr_str.replace('nan', 'x')
        arr_str = arr_str.replace('.', '')

        ma = MadArray(self.x_int, self.m)
        n_miss = np.count_nonzero(self.m)
        string = 'MadArray, dtype=int64, {} missing entries (50.0%)\n{}'
        assert str(ma) == string.format(n_miss, arr_str)

        string = '<MadArray at {}>'
        assert repr(ma) == string.format(hex(id(ma)))

    def test_merge_masks(self):
        ma1 = MadArray(self.x_float, self.m)
        ma2 = MadArray(self.x_float, self.m2)

        output = _merge_masks(ma1, ma2)

        assert isinstance(output, dict)
        assert 'mask' in output
        np.testing.assert_equal(output['mask'], np.logical_or(self.m, self.m2))

        ma1 = MadArray(self.x_float, self.m)
        ma2 = MadArray(self.x_complex, mask_magnitude=self.mm,
                       mask_phase=self.mp, complex_masking=True)

        output = _merge_masks(ma1, ma2)

        assert isinstance(output, dict)
        assert 'mask_magnitude' in output
        assert 'mask_phase' in output
        np.testing.assert_equal(output['mask_magnitude'],
                                np.logical_or(self.m, self.mm))

        np.testing.assert_equal(output['mask_phase'],
                                np.logical_or(self.m, self.mp))

        output = _merge_masks(ma2, ma1)

        assert isinstance(output, dict)
        assert 'mask_magnitude' in output
        assert 'mask_phase' in output
        np.testing.assert_equal(output['mask_magnitude'],
                                np.logical_or(self.m, self.mm))

        np.testing.assert_equal(output['mask_phase'],
                                np.logical_or(self.m, self.mp))



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Definition of a masked array.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
"""
import warnings

import numpy as np
import resampy
import simpleaudio as sa

from scipy.io import wavfile
from scipy.signal import hilbert
import matplotlib.pyplot as plt

from .mad_array import MadArray


VALID_IO_FS = {1, 8000, 16000, 32000, 48000, 11025, 22050, 44100, 88200}


class Waveform(MadArray):
    """Subclass of MadArray to handle mono and stereo audio signals.

    :class:`Waveform` inherits from :class:`MadArray` and adds an attribute
    `fs` to store the sampling frequency, as well as methods to facilitate the
    manipulation of audio files.

    .. _type_entry_waveform:

    **Type of entries**: audio data entries can have different types, that are
    associated with specific constraints on the values:

    * *float*: the values are float between -1 and 1;
    * *int*: the values are integers between a range that depends on the
      precision.
    * *complex*: the real and imaginary parts are float between -1 and 1.

    High-precision types (float128, int64, complex256) may lead to problems and
    should be used with caution.

    The casting of a waveform in a different dtype depends on the current dtype
    and the desired dtype (complex casting is similar to float casting):

    * *Integer-to-real* casting is performed by applying on each entry
      :math:`x` the function :math:`f(x)=\\frac{x - z}{2^{n-1}}`, where the
      source integral type is coded with :math:`n` bits, and :math:`z` is the
      integer associated with zero, i.e., :math:`z=0` for a signed type (`int`)
      and :math:`z=2^{n-1}` for an unsigned type (`uint`).
    * *Real-to-integer* casting is performed by applying on each entry
      :math:`x` the function :math:`f(x)=\\lfloor\\left(x + 1\\right) 2^{n-1} +
      m\\rfloor`, where the target integral type is coded with :math:`n` bits,
      and :math:`m` is the minimum integer value, i.e., :math:`m=-2^{n-1}` for
      a signed type (`int`) and :math:`z=0` for an unsigned type (`uint`);
    * *Real-to-real* casting is obtained by a basic rounding operation;
    * *Integer-to-integer* casting is obtained by chaining an
      integer-to-float64 casting and a float64-to-integer casting.

    Clipping is performed for unexpected values:

    * When casting to `float`, values outside :math:`[-1, 1]` are clipped;
    * When casting to `int`, values outside the minimum and maximum values
      allowed by the integral type are clipped:

        * :math:`\\left[-2^{n-1}, 2^{n-1}-1\\right]` for :math:`n`-bits signed
          integers;
        * :math:`\\left[0, 2^{n}-1\\right]` for :math:`n`-bits unsigned
          integers.

    These constraints are only applied when calling explicitely the method
    :meth:`astype`.

    ..  _masking:

    **Masking**: Waveform allows for complex entries, but only real-like
    masking is permitted, i.e. it is not possible to mask only the phase or the
    amplitude. In particular, this implies that the attribute
    :attr:`_is_complex` inherited from :class:`MadArray` is always equal to
    False.

    Parameters
    ----------
    data : nd-array [N] or [N, 2]
        Audio samples, as a N-length vector for a mono signal or a
        [N, 2]-shape array for a stereo signal
    fs : int, optional
        Sampling frequency of the original signal, in Hz. If float, truncated.
        If None and :paramref:`data` is a Waveform, use `data.fs`, otherwise it
        is set to 1.
    mask : nd-array, optional
        Boolean mask with True values for missing samples. Its shape must be
        the same as :paramref:`data`.
    indexing :
        See :class:`MadArray`.
    """

    def __new__(cls, data, mask=None, fs=None,  masked_indexing=None):
        if fs is None:
            if isinstance(data, Waveform):
                fs = data.fs
            else:
                fs = 1

        # create the MadArray
        data = data.squeeze()
        obj = MadArray.__new__(cls=cls, data=data, mask=mask,
                               masked_indexing=masked_indexing,
                               complex_masking=False)

        if obj.ndim > 2 or obj.ndim == 2 and obj.shape[1] > 2:
            raise ValueError('`data` should be either mono or stereo.')

        # add the new attribute to the created instance
        obj.fs = int(fs)

        return obj

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self._fs = getattr(obj, '_fs', 1)
        self._play_object = getattr(obj, '_play_object', None)

    def __reduce__(self):
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._fs, )
        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._play_object = None
        self._fs = state[-1]
        super().__setstate__(state[0:-1])

    @property
    def fs(self):
        """Frequency sampling of the audio signal.

        The signal is not resampled when the sampling frequency is modified.

        Raises
        ------
        ValueError
            If `fs` is not positive.
        """
        return self._fs

    @fs.setter
    def fs(self, fs):
        if fs <= 0:
            errmsg = 'fs is not strictly positive (given: {})'
            raise ValueError(errmsg.format(fs))
        self._fs = int(fs)

    @property
    def rms(self):
        """Root mean square error of the masked signal.

        Returns
        -------
        float

        Raises
        ------
        NotImplementedError
            If data is with an integer type.
        """
        if np.issubdtype(self.dtype, np.integer):
            errmsg = 'RMS is not available for integer data types'
            raise NotImplementedError(errmsg)

        return np.sqrt(np.mean(np.abs(self.to_np_array(0))**2))

    @property
    def length(self):
        """Length of the signal, in samples."""
        return self.shape[0]

    @property
    def duration(self):
        """Duration of the signal, in seconds."""
        return self.length / self.fs

    @property
    def n_channels(self):
        """Number of channels (1 for mono, 2 for stereo)."""
        return 1 if self.ndim == 1 else self.shape[1]

    @property
    def time_axis(self):
        """Time axis."""
        return np.arange(self.length) / self.fs

    def is_stereo(self):
        """Indicates whether the signal is stereo or not."""
        return self.n_channels == 2

    def set_rms(self, rms):
        """Set root mean square error of the signal.

        Parameters
        ----------
        rms : float
            Root mean square of the signal.

        Raises
        ------
        ValueError
            If `rms` is negative.
        NotImplementedError
            If data is with an integer type.
        """
        if rms < 0:
            errmsg = '`rms` should be a positive float (given: {})'
            raise ValueError(errmsg.format(rms))

        self[:] /= self.rms
        self[:] *= rms

    def resample(self, fs):
        """Resample the audio signal, in place.

        Can be only performed on a waveform without missing data.

        Parameters
        ----------
        fs : int
            New sampling frequency.

        Raises
        ------
        ValueError
            If `fs` is not a positive integer.
            If `self` has missing samples.
        """
        if fs <= 0:
            errmsg = '`fs` should be a positive integer (given: {})'
            raise ValueError(errmsg.format(fs))

        if self.is_masked():
            errmsg = 'Waveform has missing entries.'
            raise ValueError(errmsg)

        if fs != self.fs:
            x = self.to_np_array()

            y = resampy.resample(x, self.fs, fs, axis=0)

            self.resize(y.shape, refcheck=False)
            self[:] = y
            self._fs = fs
            self._mask = np.zeros_like(self, dtype=np.bool)

    def plot(self, x_axis_label='Time (s)', y_axis_label='s(t)',
             cpx_mode='real', fill_value=0, **kwargs):
        """Plot the signal.

        Parameters
        ----------
        x_axis_label : str
            Label of the x axis.
        y_axis_label : str
            Label of the y axis.
        cpx_mode : {'real', 'imag', 'both'}
            In case of a complex-valued signal, plot the real part only
            ('real'), the imaginary part only ('imag'), or both curves
            ('both'). This option has no effect if the signal is real-valued.
        fill_value : scalar or None
            Value used to fill missing data for display. If None, the existing
            value is used (e.g., for clipping).

        Other Parameters
        ----------------
        kwargs
            See :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        List of Line
            See :func:`matplotlib.pyplot.plot`.
        """
        x = self.to_np_array(fill_value)

        if np.issubdtype(x.dtype, np.complexfloating):
            if cpx_mode == 'real':
                lines = plt.plot(self.time_axis, np.real(x), **kwargs)
            elif cpx_mode == 'imag':
                lines = plt.plot(self.time_axis, np.imag(x), **kwargs)
            elif cpx_mode == 'both':
                lines_real = plt.plot(self.time_axis, np.real(x), **kwargs)
                lines_complex = plt.plot(
                    self.time_axis, np.imag(x), **kwargs)
                lines = lines_real + lines_complex
            else:
                errmsg = 'Unknown complex mode (given: {}).'
                raise ValueError(errmsg.format(cpx_mode))
        else:
            lines = plt.plot(self.time_axis, x, **kwargs)

        # add label_dict
        if x_axis_label is not None:
            plt.xlabel(x_axis_label)
        if y_axis_label is not None:
            plt.ylabel(y_axis_label)

        return lines

    def plot_mask(self,
                  x_axis_label='Time (s)',
                  y_axis_label='Missing',
                  **kwargs):
        """Plot the mask.

        Parameters
        ----------
        x_axis_label : str
            Label of the time axis (horizontal axis).
        y_axis_label : str
            Label of the frequency axis (vertical axis).
        """
        plt.step(
            self.time_axis, self._mask, linewidth=2, where='pre', **kwargs)
        plt.ylim(-0.25, 1.25)
        plt.yticks([0, 1])

        plt.xlabel(x_axis_label)
        plt.ylabel(y_axis_label)
        plt.grid()

    def to_wavfile(self, filename, dtype=None):
        """Save the wavefile as an audio wav file.

        Parameters
        ----------
        filename : str or :class:`pathlib.Path`
            Name of the file including paths and extension
        dtype : data-type or None, optional
            Type of the entries of the wav file. If None, use current dtype of
            data.

        Raises
        ------
        UserWarning
            If the signal is complex.
        NotImplementedError
            If dtype is not supported by the current implementation.

        See also
        --------
        scipy.io.wavfile.write
        """
        if int(self.fs) not in VALID_IO_FS:
            errmsg = '`fs` is not a valid sampling frequency (given: {}).'
            raise ValueError(errmsg.format(self.fs))

        if dtype is None:
            dtype = self.dtype

        y = self.to_np_array(0)

        if np.iscomplexobj(y):
            warnmsg = 'Complex values have been stored by keeping only the '\
                'real part, the imaginary part has been discarded.'
            warnings.warn(warnmsg)

            y = np.real(y)

            if np.issubdtype(dtype, np.complexfloating):
                dtype = np.float32

        if dtype == np.float64:
            dtype = np.float32
        elif dtype == np.int64:
            raise NotImplementedError

        # catch clipping error
        y = y.astype(dtype=dtype)

        if not np.all(np.isfinite(y)):
            errmsg = 'Data should be finite.'
            raise ValueError(errmsg)

        wavfile.write(str(filename), data=y, rate=self.fs)

    @staticmethod
    def from_wavfile(filename, dtype=np.float64, conversion_to_mono=None):
        """Load an audio file and return an Waveform object.

        Parameters
        ----------
        filename : str or :class:`pathlib.Path`
            Name of the audio file
        dtype : data-type, optional
            Output data type. If None, the data type from the wavfile is kept.
            See :ref:`Type of Entry<type_entry_waveform>`.
        conversion_to_mono : {'left', 'right', 'mean'} or None, optional
            If None (default), no conversion is performed. If str, select the
            appropriate channel (left or right), or the mean between the left
            and right channels if stereo.

        Returns
        -------
        :class:`Waveform`
        """
        fs, x = wavfile.read(str(filename))

        if dtype is None:
            dtype = x.dtype

        if x.ndim > 1:
            if conversion_to_mono is not None:

                if conversion_to_mono == 'left':
                    x = x[:, 0]

                elif conversion_to_mono == 'right':
                    x = x[:, 1]

                elif conversion_to_mono == 'mean':
                    x = np.mean(x, axis=1)

                else:
                    errmsg = '`conversion_to_mono` in invalid (given: {}).'
                    raise ValueError(errmsg.format(conversion_to_mono))

        return Waveform(x, fs=fs).astype(dtype=dtype)

    def fade(self, mode='both', fade_duration=None, fade_length=None):
        """Apply fade-in and/or fade-out.

        Half a hanning window with the specified length or duration is used
        (exactly one parameter `fade_duration` or `fade_length` must be set).

        Parameters
        ----------
        mode : {'both', 'in', 'out'}, optional
            Signal to be processed.
        fade_duration : float, optional
            Duration, in seconds, of the fade effect.
        fade_length : int, optional
            Length, in samples, of the fade effect.

        Raises
        ------
        ValueError
            If neither `fade_duration` nor `fade_length` is given.
            If both `fade_duration` and `fade_length` are given.
            If `fade_duration` or `fade_length` are greater than the
            duration or length of the waveform.
            If `fade_duration` or `fade_length` are negative.
            If `mode` is invalid.
        """
        if fade_duration is None and fade_length is None:
            errmsg = 'Either `fade_duration` or `fade_length` must be given.'
            raise ValueError(errmsg)

        if fade_duration is not None and fade_length is not None:
            errmsg = '`fade_duration` and `fade_length` cannot be given '\
                'simultaneously.'
            raise ValueError(errmsg)

        if mode not in ['both', 'in', 'out']:
            errmsg = 'Mode is unknown: {}.'
            raise ValueError(errmsg.format(mode))

        if fade_length is None:
            fade_length = int(np.round(fade_duration * self.fs))

        if fade_length == 0:
            return

        if fade_length < 0:
            errmsg = 'Fade length cannot be negative.'
            raise ValueError(errmsg)

        if fade_length > self.shape[0]:
            errmsg = 'Fade length ({}) is greater than data length ({})'
            raise ValueError(errmsg.format(fade_length, self.shape[0]))

        w = np.sin(np.pi / 2 / fade_length * np.arange(fade_length))**2

        if mode in ['both', 'in']:
            self[:len(w)] *= w

        if mode in ['both', 'out']:
            self[-len(w):] *= w[-1::-1]

    def show_player(self):
        """Show the player when using jupyter notebook."""
        try:
            from IPython.display import Audio
            if self.fs == 1:
                errmsg = 'Invalid sampling frequency: {}Hz'
                raise ValueError(errmsg.format(self.fs))

            if np.iscomplexobj(self):
                warnmsg = 'Only the real part is played.'
                warnings.warn(warnmsg)

            x = np.real(self).to_np_array(0)
            return Audio(data=x, rate=self.fs)
        except ImportError:
            print('This method should only be called in notebooks.')

    def play(self, scale=False):
        """Play the audio file.

        Parameters
        ----------
        scale : bool
            If True, the data is normalized by maximum value.

        Returns
        -------
        `simpleaudio.PlayObject`
            Object useful to handle the played audio stream.
        """

        if self.fs == 1:
            errmsg = 'Invalid sampling frequency: {}Hz'
            raise ValueError(errmsg.format(self.fs))

        if np.iscomplexobj(self):
            warnmsg = 'Only the real part is played.'
            warnings.warn(warnmsg)

        x = np.real(self)

        if scale:
            x = x.astype(np.float64)
            maxima = np.max(np.abs(self))
            if maxima > 0:
                x[:] = x / maxima * (1 - np.finfo(np.float64).eps)

        x = x.astype(dtype=np.int16).to_np_array(0)
        self._play_object = sa.play_buffer(x, num_channels=self.n_channels,
                                           bytes_per_sample=2,
                                           sample_rate=self.fs)

    def stop(self):
        """Stop playing (does nothing if the sound is not played) """
        if self.is_playing():
            self._play_object.stop()

    def is_playing(self):
        """Indicates whether the sound is currently being played.

        Returns
        -------
        bool
            True if the sound is currently being played, False otherwise.
        """
        return (hasattr(self._play_object, 'is_playing') and
                self._play_object.is_playing())

    def get_analytic_signal(self, N=None, axis=0):
        """Return the analytic waveform.

        Parameters
        ----------
        N : int, optional
            Number of Fourier components. If None, set to x.shape[axis].
        axis : int, optional
            Axis along which to do the transformation.

        Returns
        -------
        Waveform

        Raises
        ------
        TypeError
            If Waveform has missing samples.

        See Also
        --------
        scipy.signal.hilbert
        """
        if np.any(self.unknown_mask):
            errmsg = 'Waveform has missing samples.'
            raise TypeError(errmsg)

        y = hilbert(self, N=N, axis=axis)
        return Waveform(data=y, fs=self.fs)

    def clip(self, min_value=None, max_value=None):
        """Clip audio data according to given bound values.

        See :ref:`Type of Entry<type_entry_waveform>` for details.

        Parameters
        ----------
        max_value : float, optional
            Clipping is applied above this value.
        min_value : float, optional
            Clipping is applied below this value.

        Raises
        ------
        UserWarning
            If some coefficients are actually clipped.

        Notes
        -----
        Masked values can be also clipped.
        """
        if min_value is None:
            min_value = np.min(self)

        if max_value is None:
            max_value = np.max(self)

        if np.issubdtype(self.dtype, np.complexfloating):
            w_real = Waveform(np.real(self))
            w_imag = Waveform(np.imag(self))

            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                w_real.clip(min_value=min_value, max_value=max_value)

            warnmsg = 'Real part of the complex entries: {}'
            for w_i in w:
                warnings.warn(warnmsg.format(str(w_i.message)))

            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                w_imag.clip(min_value=min_value, max_value=max_value)

            warnmsg = 'Imaginary part of the complex entries: {}'
            for w_i in w:
                warnings.warn(warnmsg.format(str(w_i.message)))

            self[:] = np.array(w_real) + 1j * np.array(w_imag)

        else:
            ind_min = self < min_value
            if np.any(ind_min):
                self[ind_min] = min_value
                warnmsg = '{} values lower than {} have been clipped.'
                warnings.warn(warnmsg.format(self.dtype, min_value))

            ind_max = self > max_value
            if np.any(ind_max):
                self[ind_max] = max_value
                warnmsg = '{} values greater than {} have been clipped.'
                warnings.warn(warnmsg.format(self.dtype, max_value))

    def astype(self, dtype):
        """Cast the waveform into a given type (float, int, or complex).

        Parameters
        ----------
        dtype : data-type, optional
            Output data type. See :ref:`Type of Entry<type_entry_waveform>`.

        Returns
        -------
        nd-array
            Waveform with the desired dtype.

        Raises
        ------
        TypeError
            If argument `dtype` is an unsupported data type.
        UserWarning
            If some coefficients are actually clipped.
        """
        y = self.copy()

        if (np.issubdtype(self.dtype, np.floating) or
                np.issubdtype(self.dtype, np.complexfloating)):

            if (np.issubdtype(dtype, np.floating) or
                    np.issubdtype(dtype, np.complexfloating)):
                y.clip(min_value=-1, max_value=(1 - np.finfo(dtype).eps))
                y = np.array(y).astype(dtype)

            elif np.issubdtype(dtype, np.integer):
                target_type_info = np.iinfo(dtype)
                int_range = target_type_info.max - target_type_info.min + 1
                y = np.floor((np.real(y) + 1) / 2 * int_range +
                             target_type_info.min)
                y.clip(min_value=np.iinfo(dtype).min,
                       max_value=np.iinfo(dtype).max)
                y = np.array(y).astype(dtype)

            else:
                errmsg = 'Unsupported target type {}.'
                raise TypeError(errmsg.format(dtype))

        else:
            src_iinfo = np.iinfo(y.dtype)

            if np.issubdtype(dtype, np.floating):
                int_range = src_iinfo.max - src_iinfo.min + 1
                zero = src_iinfo.min + int_range // 2
                y = (y - zero) / (int_range // 2)
                y.clip(min_value=-1, max_value=1-np.finfo(dtype).eps)
                y = y.astype(dtype)

            elif np.issubdtype(dtype, np.integer):
                y = y.astype(dtype=np.float64)
                y = y.astype(dtype=dtype)

            elif np.issubdtype(dtype, np.complexfloating):
                y = y.astype(dtype=np.float64)
                y = np.array(y).astype(dtype=dtype)

            else:
                errmsg = 'Unsupported target type {}.'
                raise TypeError(errmsg.format(dtype))

        return Waveform(y, mask=self.unknown_mask, fs=self.fs)

    def copy(self):
        return Waveform(self)

    def __add__(self, other):
        if isinstance(other, Waveform) and self.fs != other.fs:
            errmsg = 'Waveforms do not have the same fs: {} and {}'
            raise ValueError(errmsg.format(self.fs, other.fs))

        return super().__add__(other)

    def __sub__(self, ma):
        return self.__add__(-ma)

    def __mul__(self, other):
        if isinstance(other, Waveform) and self.fs != other.fs:
            errmsg = 'Waveforms do not have the same fs: {} and {}'
            raise ValueError(errmsg.format(self.fs, other.fs))

        return super().__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Waveform) and self.fs != other.fs:
            errmsg = 'Waveforms do not have the same fs: {} and {}'
            raise ValueError(errmsg.format(self.fs, other.fs))

        return super().__truediv__(other)

    def __floordiv__(self, other):
        if isinstance(other, Waveform) and self.fs != other.fs:
            errmsg = 'Waveforms do not have the same fs: {} and {}'
            raise ValueError(errmsg.format(self.fs, other.fs))

        return super().__floordiv__(other)

    def __eq__(self, other):
        if isinstance(other, Waveform) and self.fs != other.fs:
            errmsg = 'Waveforms do not have the same fs: {} and {}'
            raise ValueError(errmsg.format(self.fs, other.fs))

        return super().__eq__(other)

    def is_equal(self, other):
        return self.fs == other.fs and super().is_equal(other)

    def __str__(self):

        string = super().__str__()
        string = string.replace('MadArray', 'Waveform, fs={}Hz, length={}')

        return string.format(self.fs, self.length)

    def __repr__(self):
        string = super().__repr__()
        string = string.replace('MadArray', 'Waveform')
        return string



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Test of the module :mod:`waveform`.

NOTE: This module uses random generation for some tests, the use of
pytest-randomly is recommanded to manage the seeds used in these tests.

TODO: handle warnings with `astype` methods (currenly, all warnings are caught
using 'warnings.catch_warnings(record=True)').

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
"""
import pickle
import pytest
import tempfile
import unittest.mock as mock
import warnings


import simpleaudio as sa

import matplotlib.pyplot as plt
import numpy as np
from scipy.io import wavfile
from IPython.display import Audio

from madarrays.waveform import Waveform

from .utils import assert_array_less_or_equal
from .utils import generate_mask_50

FLT_DTYPES = [np.float16, np.float32, np.float64]
INT_DTYPES = [np.uint8, np.int16, np.int32]
CPX_DTYPES = [np.complex64, np.complex128]

FLT_EPS = {s: np.finfo(s).eps for s in FLT_DTYPES}
CPX_EPS = {s: np.finfo(s).eps for s in CPX_DTYPES}
INT_MAX = {s: np.iinfo(s).max for s in INT_DTYPES}
INT_MIN = {s: np.iinfo(s).min for s in INT_DTYPES}
INT_ZERO = {s: (np.iinfo(s).max + np.iinfo(s).min + 1) / 2 for s in INT_DTYPES}

FS = [8000, 16000, 32000, 48000, 11025, 22050, 44100, 88200]


def assert_clipping(x_clipped, x_ref, min_value, max_value):
    """Test clipping of a referenced signal given min and max values."""

    ind_clipped_max = x_ref > max_value
    ind_clipped_min = x_ref < min_value
    ind_nonclipped = np.logical_and(~ind_clipped_max, ~ind_clipped_min)

    err_msg = 'Large values have not been clipped correctly'
    np.testing.assert_almost_equal(x_clipped[ind_clipped_max],
                                   max_value,
                                   err_msg=err_msg)

    err_msg = 'Small values have not been clipped correctly'
    np.testing.assert_almost_equal(x_clipped[ind_clipped_min],
                                   min_value, err_msg=err_msg)

    err_msg = 'Non-clipped values have been modified'
    np.testing.assert_almost_equal(x_clipped[ind_nonclipped],
                                   x_ref[ind_nonclipped], err_msg=err_msg)


@pytest.fixture(scope='class')
def get_data(request):

    request.cls.fs = np.random.choice(FS)
    request.cls.length = np.random.randint(request.cls.fs, request.cls.fs * 2)

    request.cls.x_mono = np.random.rand(request.cls.length)
    request.cls.x_mono_2 = np.random.rand(request.cls.length, 1)
    request.cls.m_mono = generate_mask_50(request.cls.length)

    request.cls.x_stereo = np.random.rand(request.cls.length, 2)
    request.cls.m_stereo = np.stack((generate_mask_50(request.cls.length),
                                     generate_mask_50(request.cls.length)),
                                    axis=-1)

    # Toy signals
    request.cls.w_float_ref = {s: Waveform(np.array(
        [-1] * 4 + [-0.5, 0, 0.5] + [1 - FLT_EPS[s]] * 4, dtype=s), fs=1)
        for s in FLT_DTYPES}

    request.cls.w_int_ref = {s: Waveform(np.array(
        [INT_MIN[s]] * 4 +
        [(INT_MIN[s] + INT_ZERO[s] + 1) // 2] +
        [INT_ZERO[s]] +
        [(INT_MAX[s] + INT_ZERO[s] + 1) // 2] +
        [INT_MAX[s]] * 4, dtype=s), fs=1) for s in INT_DTYPES}

    request.cls.w_complex_ref = {s: Waveform((1 + 1j) * np.array(
        [-1] * 4 + [-0.5, 0, 0.5] + [1 - CPX_EPS[s]] * 4,
        dtype=s), fs=1) for s in CPX_DTYPES}

    request.cls.w_float_clip = {s: Waveform(np.array(
        [-100, -2, -1.1, -1, -0.5, 0, 0.5, 1, 1, 2, 100],
        dtype=s), fs=1) for s in FLT_DTYPES}

    request.cls.w_complex_clip = {s: Waveform((1 + 1j) * np.array(
        [-100, -2, -1.1, -1, -0.5, 0, 0.5, 1, 1.1, 2, 100], dtype=s), fs=1)
        for s in CPX_DTYPES}


@pytest.mark.usefixtures('get_data')
class TestWaveform:

    def test_init_mono(self):

        w = Waveform(self.x_mono, self.m_mono)

        assert w.fs == 1
        assert w.length == self.length
        assert w.duration == self.length
        assert not w.is_stereo()
        np.testing.assert_equal(w.time_axis, np.arange(self.length))

        w = Waveform(data=self.x_mono, mask=self.m_mono, fs=self.fs)

        assert w.fs == self.fs
        assert w.length == self.length
        assert w.duration == self.length / self.fs
        np.testing.assert_equal(w.time_axis, np.arange(self.length) / self.fs)

        w = Waveform(data=self.x_mono_2, mask=self.m_mono, fs=self.fs)

        assert w.fs == self.fs
        assert w.length == self.length
        assert w.duration == self.length / self.fs
        np.testing.assert_equal(w.time_axis, np.arange(self.length) / self.fs)

    def test_init_stereo(self):

        w = Waveform(self.x_stereo, self.m_stereo)

        assert w.fs == 1
        assert w.length == self.length
        assert w.duration == self.length
        assert w.is_stereo()
        np.testing.assert_equal(w.time_axis, np.arange(self.length))

        w = Waveform(data=self.x_stereo, mask=self.m_stereo, fs=self.fs)

        assert w.fs == self.fs
        assert w.length == self.length
        assert w.duration == self.length / self.fs
        np.testing.assert_equal(w.time_axis, np.arange(self.length) / self.fs)

    def test_init_raises(self):

        # Non-stereo
        with pytest.raises(ValueError,
                           match='`data` should be either mono or stereo.'):
            Waveform(np.random.rand(100, 1000), fs=44100)

        # Non 1D or 2D array
        with pytest.raises(ValueError,
                           match='`data` should be either mono or stereo.'):
            Waveform(np.random.rand(2, 100, 1000), fs=44100)

        # Negative frequency sampling
        match = 'fs is not strictly positive \(given: -\d+\)'
        with pytest.raises(ValueError, match=match):
            Waveform(self.x_mono, fs=-self.fs)

        # Frequency sampling equal to 0
        with pytest.raises(ValueError,
                           match='fs is not strictly positive \(given: 0\)'):
            Waveform(self.x_mono, fs=0)

    def test_init_from_waveform(self):

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)
        w2 = Waveform(w)
        assert w2.fs == self.fs

    def test_resample(self):

        for fs in FS:
            # Mono
            w = Waveform(self.x_mono, fs=self.fs)
            w.resample(fs=fs)

            assert w.fs == fs
            assert w.length == int(np.floor(fs * self.length / self.fs))
            assert not w.is_stereo()
            assert w.duration == pytest.approx(self.length / self.fs, 1e-4)
            np.testing.assert_equal(
                w.time_axis,
                np.arange(np.floor(fs * self.length / self.fs)) / fs)

            # Stereo
            w = Waveform(self.x_stereo, fs=self.fs)
            w.resample(fs=fs)

            assert w.fs == fs
            assert w.length == int(np.floor(fs * self.length / self.fs))
            assert w.is_stereo()
            assert w.duration == pytest.approx(self.length / self.fs, 1e-4)
            np.testing.assert_equal(
                w.time_axis,
                np.arange(np.floor(fs * self.length / self.fs)) / fs)

        # Negative frequency sampling
        with pytest.raises(
                ValueError,
                match='`fs` should be a positive integer \(given: -\d+\)'):
            w.resample(fs=-self.fs)

        # Frequency sampling equal to 0
        with pytest.raises(
                ValueError,
                match='`fs` should be a positive integer \(given: 0\)'):
            w.resample(fs=0)

        # Masked data
        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)
        with pytest.raises(ValueError, match='Waveform has missing entries.'):
            w.resample(fs=fs)

    def test_set_rms(self):

        # Float
        w = Waveform(self.x_mono, fs=self.fs)
        assert w.rms == pytest.approx(np.sqrt(1 / 3), 1e-1)

        w.set_rms(1)
        assert w.rms == pytest.approx(1, 1e-7)
        assert np.sqrt(np.mean(w**2)) == pytest.approx(1, 1e-7)

        # Negative rms
        with pytest.raises(
                ValueError,
                match='`rms` should be a positive float \(given: -1\)'):
            w.set_rms(-1)

        # Int
        w = Waveform(self.x_mono, fs=self.fs)
        for w_int in self.w_int_ref.values():
            # Getting RMS
            with pytest.raises(
                    NotImplementedError,
                    match='RMS is not available for integer data types'):
                w_int.rms

    def test_clip_float(self):

        min_value = -np.random.rand() * np.random.randint(100)
        max_value = np.random.rand() * np.random.randint(100)

        x_noclip = np.linspace(min_value, max_value, 1000)
        x_below_clip = np.linspace(min_value - 1, max_value, 1000)
        x_above_clip = np.linspace(min_value, max_value + 1, 1000)

        # No clipping
        w = Waveform(x_noclip)
        w.clip(max_value=max_value, min_value=min_value)

        assert_clipping(x_clipped=w, x_ref=x_noclip,
                        min_value=min_value, max_value=max_value)

        # Clipping from below
        w = Waveform(x_below_clip)
        match = 'float\d+ values lower than -\d+.\d+ have been clipped.'
        with pytest.warns(UserWarning, match=match):
            w.clip(min_value=min_value)
        assert_clipping(x_clipped=w, x_ref=x_below_clip,
                        min_value=min_value, max_value=max_value)

        # Clipping by above
        w = Waveform(x_above_clip)
        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        with pytest.warns(UserWarning, match=match):
            w.clip(max_value=max_value)

        assert_clipping(x_clipped=w, x_ref=x_above_clip,
                        min_value=min_value, max_value=max_value)

    def test_clip_complex(self):

        min_value = -np.random.rand() * np.random.randint(100)
        max_value = np.random.rand() * np.random.randint(100)

        x_noclip = (np.linspace(min_value, max_value, 1000) +
                    1j * np.linspace(min_value, max_value, 1000))
        x_real_below_clip = (np.linspace(min_value - 1, max_value, 1000) +
                             1j * np.linspace(min_value, max_value, 1000))
        x_imag_above_clip = (np.linspace(min_value, max_value, 1000) +
                             1j * np.linspace(min_value, max_value + 1, 1000))

        # No clipping
        w = Waveform(x_noclip)
        w.clip(max_value=max_value, min_value=min_value)

        assert_clipping(x_clipped=np.real(w), x_ref=np.real(x_noclip),
                        min_value=min_value, max_value=max_value)
        assert_clipping(x_clipped=np.imag(w), x_ref=np.imag(x_noclip),
                        min_value=min_value, max_value=max_value)

        # Clipping from below of the real part
        match = 'Real part of the complex entries: float\d+ values lower than'\
                ' -\d+.\d+ have been clipped.'
        w = Waveform(x_real_below_clip)

        with pytest.warns(UserWarning, match=match):
            w.clip(max_value=max_value, min_value=min_value)

        assert_clipping(x_clipped=np.real(w), x_ref=np.real(x_real_below_clip),
                        min_value=min_value, max_value=max_value)
        assert_clipping(x_clipped=np.imag(w), x_ref=np.imag(x_real_below_clip),
                        min_value=min_value, max_value=max_value)

        # Clipping by above of the imaginary part
        match = 'Imaginary part of the complex entries: float\d+ values '\
                'greater than \d+.\d+ have been clipped.'
        w = Waveform(x_imag_above_clip)

        with pytest.warns(UserWarning, match=match):
            w.clip(max_value=max_value, min_value=min_value)

        assert_clipping(x_clipped=np.real(w), x_ref=np.real(x_imag_above_clip),
                        min_value=min_value, max_value=max_value)
        assert_clipping(x_clipped=np.imag(w), x_ref=np.imag(x_imag_above_clip),
                        min_value=min_value, max_value=max_value)

    def test_astype_float_float(self):

        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        for dtype_src, w_src in self.w_float_ref.items():
            for dtype_targ, w_targ in self.w_float_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))

                if FLT_EPS[dtype_src] < FLT_EPS[dtype_targ]:
                    with pytest.warns(UserWarning, match=match):
                        w_est = w_src.astype(dtype_targ)
                else:
                    w_est = w_src.astype(dtype_targ)

                assert w_est.dtype == w_targ.dtype

                if FLT_EPS[dtype_src] <= FLT_EPS[dtype_targ]:
                    np.testing.assert_equal(w_est, w_targ)
                else:
                    np.testing.assert_allclose(
                        w_est, w_targ, atol=FLT_EPS[dtype_src], rtol=0)

        for dtype_src, w_src in self.w_float_clip.items():
            for dtype_targ, w_targ in self.w_float_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                with pytest.warns(UserWarning, match=match):
                    w_est = w_src.astype(dtype=dtype_targ)

                if FLT_EPS[dtype_src] <= FLT_EPS[dtype_targ]:
                    np.testing.assert_equal(w_est, w_targ)
                else:
                    np.testing.assert_allclose(
                        w_est, w_targ, atol=FLT_EPS[dtype_src], rtol=0)

    def test_astype_int_int(self):
        for dtype_src, w_src in self.w_int_ref.items():
            for dtype_targ, w_targ in self.w_int_ref.items():
                print("Casting from {} to {}.".format(dtype_src, dtype_targ))

                w_est = w_src.astype(dtype_targ)
                atol = ((INT_MAX[dtype_targ] - INT_MIN[dtype_targ] + 1) /
                        (INT_MAX[dtype_src] - INT_MIN[dtype_src] + 1))

                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

    def test_astype_complex_complex(self):

        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        for dtype_src, w_src in self.w_complex_ref.items():
            for dtype_targ, w_targ in self.w_complex_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))

                atol = 2 * (CPX_EPS[dtype_src] + CPX_EPS[dtype_targ])

                if CPX_EPS[dtype_src] < CPX_EPS[dtype_targ]:
                    with pytest.warns(UserWarning, match=match):
                        w_est = w_src.astype(dtype_targ)
                else:
                    w_est = w_src.astype(dtype_targ)

                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

        for dtype_src, w_src in self.w_complex_clip.items():
            for dtype_targ, w_targ in self.w_complex_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                atol = 2 * (CPX_EPS[dtype_src] + CPX_EPS[dtype_targ])
                with pytest.warns(UserWarning, match=match):
                    w_est = w_src.astype(dtype=dtype_targ)
                    np.testing.assert_allclose(
                        w_est, w_targ, atol=atol, rtol=0)

    def test_astype_float_int(self):
        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        for dtype_src, w_src in self.w_float_ref.items():
            for dtype_targ, w_targ in self.w_int_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                w_est = w_src.astype(dtype_targ)

                atol = (FLT_EPS[dtype_src] *
                        (INT_MAX[dtype_targ] - INT_MIN[dtype_targ] + 1) / 2)
                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

        for dtype_src, w_src in self.w_float_clip.items():
            for dtype_targ, w_targ in self.w_int_ref.items():
                with pytest.warns(UserWarning, match=match):
                    print('Casting from {} to {}.'.format(
                        dtype_src, dtype_targ))
                    w_est = w_src.astype(dtype=dtype_targ)
                    np.testing.assert_array_equal(w_est, w_targ)

    def test_astype_int_float(self):
        for dtype_src, w_src in self.w_int_ref.items():
            for dtype_targ, w_targ in self.w_float_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))

                with warnings.catch_warnings(record=True):
                    w_est = w_src.astype(dtype_targ)

                atol = ((INT_MAX[dtype_src] - INT_MIN[dtype_src] + 1) +
                        FLT_EPS[dtype_targ] / 2)
                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

    def test_astype_float_complex(self):
        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        for dtype_src, w_src in self.w_float_ref.items():
            for dtype_targ, w_targ in self.w_complex_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                atol = FLT_EPS[dtype_src] + CPX_EPS[dtype_targ]

                if FLT_EPS[dtype_src] < CPX_EPS[dtype_targ]:
                    with pytest.warns(UserWarning, match=match):
                        w_est = w_src.astype(dtype_targ)
                else:
                    w_est = w_src.astype(dtype_targ)

                np.testing.assert_allclose(w_est, np.real(w_targ),
                                           atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

        for dtype_src, w_src in self.w_float_clip.items():
            for dtype_targ, w_targ in self.w_complex_ref.items():
                msg = "Casting from {} to {}.".format(dtype_src, dtype_targ)
                atol = FLT_EPS[dtype_src] + CPX_EPS[dtype_targ]
                with pytest.warns(UserWarning, match=match):
                    w_est = w_src.astype(dtype=dtype_targ)
                    np.testing.assert_allclose(w_est,
                                               np.real(w_targ).astype(
                                                   dtype_targ),
                                               atol=atol, rtol=0, err_msg=msg)

    def test_astype_complex_float(self):
        match = 'float\d+ values greater than \d+.\d+ have been clipped.'
        for dtype_src, w_src in self.w_complex_ref.items():
            for dtype_targ, w_targ in self.w_float_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                atol = CPX_EPS[dtype_src] + FLT_EPS[dtype_targ]

                if CPX_EPS[dtype_src] < FLT_EPS[dtype_targ]:
                    with pytest.warns(UserWarning, match=match):
                        w_est = w_src.astype(dtype_targ)
                else:
                    with pytest.warns(np.ComplexWarning):
                        w_est = w_src.astype(dtype_targ)
                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

    def test_astype_complex_int(self):
        for dtype_src, w_src in self.w_complex_ref.items():
            for dtype_targ, w_targ in self.w_int_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                atol = (CPX_EPS[dtype_src] *
                        (INT_MAX[dtype_targ] - INT_MIN[dtype_targ] + 1) / 2)

                w_est = w_src.astype(dtype_targ)
                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

    def test_astype_int_complex(self):
        for dtype_src, w_src in self.w_int_ref.items():
            for dtype_targ, w_targ in self.w_complex_ref.items():
                print('Casting from {} to {}.'.format(dtype_src, dtype_targ))
                atol = ((INT_MAX[dtype_src] - INT_MIN[dtype_src] + 1) +
                        CPX_EPS[dtype_targ] / 2)

                w_est = w_src.astype(dtype_targ)

                np.testing.assert_allclose(w_est, w_targ, atol=atol, rtol=0)
                assert w_est.dtype == w_targ.dtype

    def test_astype_invalid(self):
        match = "Unsupported target type <class 'list'>."
        with pytest.raises(TypeError, match=match):
            self.w_float_ref[np.float64].astype(list)

        with pytest.raises(TypeError, match=match):
            self.w_int_ref[np.int32].astype(list)

    def test_to_wavfile_mono(self):

        match = 'Complex values have been stored by keeping only the real '\
                'part, the imaginary part has been discarded.'

        for fs in FS:
            for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
                print('Fs: {:d}, dtype: {}]'.format(fs, dtype))

                w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

                with warnings.catch_warnings(record=True):
                    w = w.astype(dtype)

                with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:

                    if dtype in CPX_DTYPES:
                        with pytest.warns(UserWarning, match=match):
                            w.to_wavfile(tmp_file.name)
                    else:
                        w.to_wavfile(tmp_file.name)

                    with warnings.catch_warnings(record=True):
                        w2 = Waveform.from_wavfile(tmp_file.name).astype(dtype)
                    np.testing.assert_almost_equal(w2.to_np_array(),
                                                   w.to_np_array(0),
                                                   decimal=3)
                    assert w.fs == w2.fs

        for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
            for dtype_targ in FLT_DTYPES:
                print('dtype: {}, dtype_targ: {}]'.format(dtype, dtype_targ))

                w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

                with warnings.catch_warnings(record=True):
                    w = w.astype(dtype)

                with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:

                    if dtype in CPX_DTYPES:
                        with pytest.warns(UserWarning, match=match):
                            w.to_wavfile(tmp_file.name, dtype=dtype_targ)
                    else:
                        with warnings.catch_warnings(record=True):
                            w.to_wavfile(tmp_file.name)

                    with warnings.catch_warnings(record=True):
                        w2 = Waveform.from_wavfile(
                            tmp_file.name).astype(dtype)
                    np.testing.assert_almost_equal(w2.to_np_array(),
                                                   w.to_np_array(0),
                                                   decimal=3)

    def test_to_wavfile_stereo(self):

        match = 'Complex values have been stored by keeping only the real '\
            'part, the imaginary part has been discarded.'
        for fs in FS:
            for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
                print('Fs: {:d}, dtype: {}]'.format(fs, dtype))

                w = Waveform(self.x_stereo, self.m_stereo, fs=self.fs)
                with warnings.catch_warnings(record=True):
                    w = w.astype(dtype)

                with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:

                    if dtype in CPX_DTYPES:
                        with pytest.warns(UserWarning, match=match):
                            w.to_wavfile(tmp_file.name)
                    else:
                        with warnings.catch_warnings(record=True):
                            w.to_wavfile(tmp_file.name)

                    with warnings.catch_warnings(record=True):
                        w2 = Waveform.from_wavfile(tmp_file.name).astype(dtype)
                    np.testing.assert_almost_equal(w2.to_np_array(),
                                                   w.to_np_array(0),
                                                   decimal=3)
                    assert w.fs == w2.fs

        for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
            for dtype_targ in FLT_DTYPES:
                print('dtype: {}, dtype_targ: {}]'.format(dtype, dtype_targ))

                w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

                with warnings.catch_warnings(record=True):
                    w = w.astype(dtype)

                with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:

                    if dtype in CPX_DTYPES:
                        with pytest.warns(UserWarning, match=match):
                            w.to_wavfile(tmp_file.name, dtype=dtype_targ)
                    else:
                        with warnings.catch_warnings(record=True):
                            w.to_wavfile(tmp_file.name)

                    with warnings.catch_warnings(record=True):
                        w2 = Waveform.from_wavfile(tmp_file.name).astype(dtype)
                    np.testing.assert_almost_equal(w2.to_np_array(),
                                                   w.to_np_array(0),
                                                   decimal=3)

    def test_to_wavfile_raises(self):
        # Unsupported/Not implemented types
        for dtype in (np.int64,):
            w = Waveform(self.x_mono.astype(dtype), fs=self.fs)
            with pytest.raises(NotImplementedError):
                w.to_wavfile('')

        # Infinite values
        w = Waveform(self.x_mono, fs=self.fs)
        for infinite_value in (np.inf, -np.inf):
            w[w.size // 2] = infinite_value
            with pytest.raises(ValueError, match='Data should be finite.'):
                w.to_wavfile('')

        # Unsupported fs
        w = Waveform(self.x_mono, fs=1234)

        match = '`fs` is not a valid sampling frequency \(given: 1234\).'
        with pytest.raises(ValueError, match=match):
            w.to_wavfile('')

    def test_from_wavfile_mono(self):

        x_len = np.random.randint(100, 1000)
        fs = np.random.choice(FS)
        x = np.random.rand(x_len)

        with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:
            wavfile.write(filename=str(tmp_file.name), data=x,
                          rate=fs)

            # Given dtype
            w_l = Waveform.from_wavfile(tmp_file.name, dtype=np.float64)
            np.testing.assert_almost_equal(w_l.data, x, decimal=4)
            assert w_l.fs == fs

            # dtype is None
            w_m = Waveform.from_wavfile(tmp_file.name, dtype=None)
            np.testing.assert_almost_equal(w_m.data, x, decimal=4)
            assert w_l.fs == fs

            # dtype is int16
            w_m.to_wavfile(str(tmp_file.name), dtype=np.int16)
            w_m_i = Waveform.from_wavfile(tmp_file.name, dtype=None)
            assert w_m_i.dtype == np.int16

    def test_from_wavfile_stereo(self):

        with tempfile.NamedTemporaryFile(suffix='.wav') as tmp_file:

            length_stereo = np.random.randint(100, 1000)
            fs_stereo = np.random.choice(FS)
            x_stereo = np.random.rand(length_stereo, 2)
            wavfile.write(filename=str(tmp_file.name), data=x_stereo,
                          rate=fs_stereo)

            # Conversion to mono is left
            w_l = Waveform.from_wavfile(tmp_file.name, dtype=np.float64,
                                        conversion_to_mono='left')
            np.testing.assert_almost_equal(x_stereo[:, 0], w_l.to_np_array(),
                                           decimal=4)
            assert w_l.fs == fs_stereo

            # Conversion to mono is right
            w_r = Waveform.from_wavfile(tmp_file.name,
                                        conversion_to_mono='right')

            np.testing.assert_almost_equal(x_stereo[:, 1],
                                           w_r.to_np_array(), decimal=4)
            assert w_l.fs == fs_stereo

            # Conversion to mono is mean
            w_m = Waveform.from_wavfile(tmp_file.name,
                                        conversion_to_mono='mean')

            np.testing.assert_almost_equal(np.mean(x_stereo, axis=1),
                                           w_m.to_np_array(), decimal=4)
            assert w_l.fs == fs_stereo

            # Stereo
            w_stereo = Waveform.from_wavfile(tmp_file.name)
            np.testing.assert_almost_equal(w_stereo.to_np_array(),
                                           x_stereo, decimal=4)
            assert w_stereo.fs == fs_stereo
            assert w_stereo.is_stereo()

            match = '`conversion_to_mono` in invalid \(given: center\).'
            with pytest.raises(ValueError, match=match):
                Waveform.from_wavfile(
                    tmp_file.name, conversion_to_mono='center')

    def test_plot_real(self):

        for x, m in zip([self.x_mono, self.x_stereo],
                        [self.m_mono, self.m_stereo]):
            w = Waveform(x, m, fs=self.fs)

            lines = w.plot(fill_value=None)
            assert type(lines[0]) == plt.matplotlib.lines.Line2D
            lines = w.plot(None, None)
            assert type(lines[0]) == plt.matplotlib.lines.Line2D
            w.plot_mask()

    def test_plot_complex(self):
        w = Waveform(self.x_mono + 1j * self.x_mono, self.m_mono, fs=self.fs)

        lines = w.plot()
        assert type(lines[0]) == plt.matplotlib.lines.Line2D
        lines = w.plot(cpx_mode='imag')
        assert type(lines[0]) == plt.matplotlib.lines.Line2D
        lines = w.plot(cpx_mode='both')
        assert type(lines[0]) == plt.matplotlib.lines.Line2D
        assert type(lines[1]) == plt.matplotlib.lines.Line2D

        # Unknows complex mode
        with pytest.raises(
                ValueError,
                match='Unknown complex mode \(given: realandimag\).'):
            w.plot(cpx_mode='realandimag')

    def test_copy(self):

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

        w_copy = w.copy()
        assert w.is_equal(w_copy)
        assert id(w) != id(w_copy)
        assert id(w._mask) != id(w_copy._mask)

    def test_str_repr(self):

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

        x = np.copy(self.x_mono)
        x[self.m_mono] = np.nan
        arr_str = np.ndarray.__str__(x)
        arr_str = arr_str.replace('nan', 'x')
        n_miss = np.count_nonzero(self.m_mono)

        exp_str = 'Waveform, fs={}Hz, length={}, dtype={}, {} missing '\
                  'entries (50.0%)\n{}'.format(self.fs, self.length, x.dtype,
                                               n_miss, arr_str)
        assert str(w) == exp_str

        exp_str = '<Waveform at {}>'
        assert repr(w) == exp_str.format(hex(id(w)))

    def test_show_player(self, capsys):

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)
        player = w.show_player()

        assert isinstance(player, Audio)

        w = Waveform(self.x_mono, self.m_mono,
                     fs=self.fs).astype(np.complex128)

        with pytest.warns(UserWarning, match='Only the real part is played.'):
            player = w.show_player()
        assert isinstance(player, Audio)

        # Wrong frequency sampling
        w = Waveform(self.x_mono, self.m_mono, fs=1)
        with pytest.raises(ValueError,
                           match='Invalid sampling frequency: \d+Hz'):
            w.show_player()

        with mock.patch.dict('sys.modules', {'IPython.display': None}):
            w.show_player()
        captured = capsys.readouterr()
        exp_print = 'This method should only be called in notebooks.\n'
        assert captured[0] == exp_print

    def test_play_stop_isplaying_mono(self):

        try:
            import simpleaudio
            simpleaudio.play_buffer(np.zeros(1000), 1, 2, 44100)
            is_patched = False
        except sa._simpleaudio.SimpleaudioError:
            is_patched = True
            patcher1 = mock.patch('simpleaudio.PlayObject', )
            patcher2 = mock.patch('simpleaudio.play_buffer')
            patcher1.start()
            patcher2.start()

        for fs in FS:
            for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
                print('Play {} data sample at {} Hz.'.format(dtype, fs))

                if np.issubdtype(dtype, np.integer):
                    w = Waveform(np.full(1000, fill_value=INT_ZERO[dtype],
                                         dtype=dtype), fs=fs)
                else:
                    w = Waveform(np.zeros(1000, dtype=dtype), fs=fs)

                assert not w.is_playing()
                w.stop()

                if np.issubdtype(dtype, np.complexfloating):
                    with pytest.warns(UserWarning,
                                      match='Only the real part is played.'):
                        w.play()
                else:
                    w.play()
                assert w.is_playing()
                w.stop()
                assert w.is_playing()

                if np.issubdtype(dtype, np.complexfloating):
                    with pytest.warns(UserWarning,
                                      match='Only the real part is played.'):
                        w.play(scale=True)
                else:
                    w.play(scale=True)

                w.stop()
                if not is_patched:
                    assert isinstance(w._play_object, sa.shiny.PlayObject)

        w = Waveform(np.zeros(1000), fs=1)

        with pytest.raises(ValueError,
                           match='Invalid sampling frequency: 1Hz'):
            w.play()

        if is_patched:
            patcher1.stop()
            patcher2.stop()

    def test_play_stop_isplaying_stereo(self):

        for fs in FS:
            for dtype in (INT_DTYPES + FLT_DTYPES + CPX_DTYPES):
                print('Play {} data sample at {} Hz.'.format(dtype, fs))

                if np.issubdtype(dtype, np.integer):
                    w = Waveform(np.full((1000, 2), fill_value=INT_ZERO[dtype],
                                         dtype=dtype), fs=fs)
                else:
                    w = Waveform(np.zeros((1000, 2), dtype=dtype), fs=fs)

                    assert not w.is_playing()
                    w.stop()

                if np.issubdtype(dtype, np.complexfloating):
                    with pytest.warns(UserWarning,
                                      match='Only the real part is played.'):
                        try:
                            w.play()
                        except sa._simpleaudio.SimpleaudioError:
                            pass
                else:
                    try:
                        w.play()
                        assert w.is_playing()
                        w.stop()
                        assert w.is_playing()
                    except sa._simpleaudio.SimpleaudioError:
                        pass

                if np.issubdtype(dtype, np.complexfloating):
                    with pytest.warns(UserWarning,
                                      match='Only the real part is played.'):
                        try:
                            w.play(scale=True)
                        except sa._simpleaudio.SimpleaudioError:
                            pass
                else:
                    try:
                        w.play(scale=True)
                        w.stop()

                        assert isinstance(w._play_object, sa.shiny.PlayObject)

                    except sa._simpleaudio.SimpleaudioError:
                        pass

        w = Waveform(np.zeros(1000), fs=1)

        with pytest.raises(ValueError,
                           match='Invalid sampling frequency: 1Hz'):
            try:
                w.play()
            except sa._simpleaudio.SimpleaudioError:
                pass

    def test_get_analytic_signal(self):

        for x, m in zip([self.x_mono, self.x_stereo],
                        [self.m_mono, self.m_stereo]):
            with pytest.raises(TypeError,
                               match='Waveform has missing samples'):
                Waveform(x, m, fs=self.fs).get_analytic_signal()

            w = Waveform(x, fs=self.fs)
            np.testing.assert_array_almost_equal(
                np.real(w.get_analytic_signal()), x)

    def test_eq_ne_numpy(self):

        for x, m in zip([self.x_mono, self.x_stereo],
                        [self.m_mono, self.m_stereo]):
            w = Waveform(x, fs=self.fs)
            cmp_w = w == x
            assert type(cmp_w) == np.ndarray
            assert cmp_w.dtype == np.bool
            assert np.all(cmp_w)

        w = Waveform(np.ones(self.length), fs=self.fs)
        cmp_w = w == 0
        assert type(cmp_w) == np.ndarray
        assert cmp_w.dtype == np.bool
        assert np.all(~cmp_w)

    def test_operations(self):
        w1 = Waveform(self.x_mono, fs=self.fs)
        w2 = Waveform(self.x_mono, fs=1 if self.fs > 1 else 44100)

        np.testing.assert_equal(w1 + w1, self.x_mono + self.x_mono,
                                err_msg='add')
        np.testing.assert_equal(w1 - w1, self.x_mono - self.x_mono,
                                err_msg='sub')
        np.testing.assert_equal(w1 * w1, self.x_mono * self.x_mono,
                                err_msg='mul')
        np.testing.assert_equal(w1 / w1, self.x_mono / self.x_mono,
                                err_msg='truediv')
        np.testing.assert_equal(w1 // w1, self.x_mono // self.x_mono,
                                err_msg='floordiv')

        match = 'Waveforms do not have the same fs: \d+ and \d+'
        with pytest.raises(ValueError, match=match):
            w1 + w2

        with pytest.raises(ValueError, match=match):
            w1 * w2

        with pytest.raises(ValueError, match=match):
            w1 / w2

        with pytest.raises(ValueError, match=match):
            w1 // w2

        with pytest.raises(ValueError, match=match):
            w1 == w2

        assert not w1.is_equal(w2)

    def test_fade(self):
        all_modes = {'both', 'in', 'out'}
        for fade_len in {self.length // 10, self.length}:
            fade_dur = fade_len / self.fs
            w = dict()
            for mode in all_modes:
                print('fade_length={}, mode={}, fs={}'.format(
                    fade_len, mode, self.fs))

                w[mode] = Waveform(self.x_mono, self.m_mono, fs=self.fs)
                w[mode].fade(mode=mode, fade_length=fade_len)
                w_with_dur = Waveform(self.x_mono, self.m_mono, fs=self.fs)
                w_with_dur.fade(mode=mode, fade_duration=fade_dur)

                np.testing.assert_array_equal(w[mode], w_with_dur)

            np.testing.assert_array_equal(
                np.array(w['both'][fade_len:-fade_len]),
                self.x_mono[fade_len:-fade_len],
                err_msg='Some samples have been modified in the center part.')

            np.testing.assert_array_equal(
                np.array(w['in'][fade_len:]),
                self.x_mono[fade_len:],
                err_msg='Samples modified after the fade-in area.')

            np.testing.assert_array_equal(
                np.array(w['out'][:-fade_len]),
                self.x_mono[:-fade_len],
                err_msg='Samples modified before the fade-out area.')

            for mode in {'both', 'in'}:
                print('fade_len={}, mode={}'.format(fade_len, mode))
                assert_array_less_or_equal(
                    np.array(w[mode][:fade_len]),
                    self.x_mono[:fade_len],
                    err_msg='Samples in the fade-in area greater.')

            for mode in {'both', 'out'}:
                print('fade_len={}, mode={}'.format(fade_len, mode))
                assert_array_less_or_equal(
                    np.array(w[mode][fade_len:]),
                    self.x_mono[fade_len:],
                    err_msg='Samples in the fade-out area  greater')

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)
        w.fade(mode='both', fade_length=0)
        np.testing.assert_array_equal(np.array(w), self.x_mono)

    def test_fade_exceptions(self):
        all_modes = {'both', 'in', 'out'}
        fade_len = self.length // 5
        fade_dur = fade_len / self.fs
        w = dict()
        for mode in all_modes:
            w[mode] = Waveform(self.x_mono, self.m_mono, fs=self.fs)
            match = 'Either `fade_duration` or `fade_length` must be given.'
            with pytest.raises(ValueError, match=match):
                w[mode].fade(mode=mode)

            match = '`fade_duration` and `fade_length` cannot be given '\
                'simultaneously.'
            with pytest.raises(ValueError, match=match):
                w[mode].fade(mode=mode,
                             fade_length=fade_len,
                             fade_duration=fade_dur)

            match = 'Fade length \(\d+\) is greater than data length \(\d+\)'
            with pytest.raises(ValueError, match=match):
                w[mode].fade(mode=mode, fade_length=self.length + 1)

            with pytest.raises(ValueError,
                               match='Fade length cannot be negative.'):
                w[mode].fade(mode=mode, fade_length=-1)

            with pytest.raises(ValueError,
                               match='Fade length cannot be negative.'):
                w[mode].fade(mode=mode, fade_duration=-1 / self.fs)

        with pytest.raises(ValueError, match='Mode is unknown: stereo.'):
            w[mode].fade(mode='stereo', fade_length=fade_len)

    def test_pickle(self):

        w = Waveform(self.x_mono, self.m_mono, fs=self.fs)

        with tempfile.NamedTemporaryFile() as tmp_file:
            with open(tmp_file.name, 'wb') as fout:
                pickle.dump(w, fout)

            with open(tmp_file.name, 'rb') as fin:
                pw = pickle.load(fin)

            assert w.is_equal(pw)
